# -*- coding: utf-8 -*-

import functools
import logging
import time
import tracemalloc


class Benchmark:

    def __init__(self, name=None, logger=None):
        self.name = name or self.__class__.__name__
        self.logger = logger or logging

    def __call__(self, fn):
        self.name = fn.__name__

        @functools.wraps(fn)
        def decorator(*args, **kwargs):
            with self:
                return fn(*args, **kwargs)

        return decorator


class Timed(Benchmark):
    """
    Usage:

    logging.basicConfig(level=logging.DEBUG)

    # Using context manager
    with Timed():
        print(sum(list(range(1000))))

    # Using decorator
    @Timed()
    def calculate():
        print(sum(list(range(1000))))
    calculate()
    """

    def __init__(self, name=None, logger=None, decimals=3):
        super().__init__(name=name, logger=logger)
        self.decimals = decimals

    def __enter__(self):
        self.start_time = time.perf_counter()

    def __exit__(self, *args, **kwargs):
        self.end_time = time.perf_counter()
        elapsed_time = self.end_time - self.start_time
        self.logger.debug(f"{self.name}: [{elapsed_time:.{self.decimals}f} seconds]")


class MemoryUsage(Benchmark):
    """
    Usage:

    logging.basicConfig(level=logging.DEBUG)

    # Using context manager
    with MemoryUsage():
        print(sum(list(range(1000))))

    # Using decorator
    @MemoryUsage()
    def calculate():
        print(sum(list(range(1000))))
    calculate()
    """

    def __enter__(self):
        if not tracemalloc.is_tracing():
            tracemalloc.start()
        tracemalloc.reset_peak()
        self.trace("starting")

    def __exit__(self, *args, **kwargs):
        self.trace("finished")
        if tracemalloc.is_tracing():
            tracemalloc.stop()

    def trace(self, when):
        this, peak = tracemalloc.get_traced_memory()
        info = "%s - %s: [Current RAM usage: %.1f KB; Peak: %.1f} KB]"
        self.logger.debug(info, self.name, when, this / 1024, peak / 1024)



# -*- coding: utf-8 -*-

"""
Useful functions for working with JSONs.
- Supports `orjson`, `ujson` libraries or standard `json`.
- Supports following compression formats: gzip => (.gz), bzip2 => (.bz2), xz => (.xz)
"""

import csv
import dataclasses
import datetime
import decimal
import functools
import itertools
import json
import operator
import os
import textwrap
import uuid
import zipfile

import pymince._constants
import pymince.file

# Use the fastest available JSON library for serialization/deserialization, prioritizing `orjson`,
# then `ujson`, and defaulting to the standard `json` if none are installed.
try:
    import orjson
except ImportError:
    orjson = None

try:
    import ujson
except ImportError:
    ujson = None

PROVIDER = orjson or ujson or json
ENCODING = pymince._constants.utf_8

json_dumps = functools.partial(PROVIDER.dumps, ensure_ascii=False)
json_dump = functools.partial(PROVIDER.dump, ensure_ascii=False)
json_load = (orjson or ujson or json).load


def load_from(filename, encoding=ENCODING):
    """
    Load JSON from a file.
    - Recognizes (`.gz`, `.xz`, `.bz2`) extensions to load compressed files.
    - Loads falls back to the functions: (`orjson.load`, `ujson.load`, and `json.load`).

    Examples:
        from pymince.json import load_from

        dictionary1 = load_from("foo.json")     # uncompressed
        dictionary2 = load_from("foo.json.gz")  # gzip-compressed
        dictionary3 = load_from("foo.json.xz")  # lzma-compressed
        dictionary4 = load_from("foo.json.bz2") # bz2-compressed
    """

    with pymince.file.xopen(filename, mode="rt", encoding=encoding) as file:
        return json.load(file)


def dump_into(filename, obj, encoding=ENCODING, **kwargs):
    """
    Dump JSON to a file.
    - Use (`.gz`, `.xz`, `.bz2`) extensions to create compressed files.
    - Dumps falls back to the functions: (`orjson.dump`, `ujson.dump`, and `json.dump`).

    Examples:
        from pymince.json import dump_into

        dump_into("foo.json", {"key": "value"})     # uncompressed
        dump_into("foo.json.gz", {"key": "value"})  # gzip-compressed
        dump_into("foo.json.xz", {"key": "value"})  # lzma-compressed
        dump_into("foo.json.bz2", {"key": "value"}) # bz2-compressed
    """

    with pymince.file.xopen(filename, mode="wt", encoding=encoding) as fd:
        json_dump(obj, fd, **kwargs)


def dump_into_zip(zip_path, arcname, payload, **kwargs):
    """
    Dump JSON into the zip archive under the name arcname.

    Examples:
        from pymince.json import dump_into_zip

        dump_into_zip("archive.zip", "foo.json", {"key": "value"})
    """

    with zipfile.ZipFile(zip_path, mode="w") as zf:
        json_string = json_dumps(payload, **kwargs)
        zf.writestr(arcname, json_string)


def load_from_zip(zip_path, arcname):
    """
    Load JSON from a file named "arcname" inside a zip archive.

    Examples:
        from pymince.json import load_from_zip

        dictionary = load_from_zip("archive.zip", "foo.json")
    """

    with zipfile.ZipFile(zip_path, mode="r") as zf, zf.open(arcname) as file:
        return json.load(file)


def dump_from_csv(
    csv_path,
    json_path,
    /,
    *,
    fieldnames=None,
    start=0,
    stop=None,
    strip=True,
    encoding=ENCODING,
    **kwargs,
):
    """
    Dump CSV file to a JSON file.
    - Use (`.gz`, `.xz`, `.bz2`) extensions to create a compressed file.
    - Dumps falls back to the functions: (`orjson.dumps`, `ujson.dumps`, and `json.dumps`).

    :param str csv_path:
    :param str json_path:
    :param fieldnames: list of keys for the JSON
    :param int start:
        If start is specified, will skip all preceding elements;
        otherwise, start defaults to zero.
    :param int stop:
    :param bool strip:
        Whether white space should be removed from the
        beginning and end of field values.
    :param str encoding: utf-8 is used by default.
    """

    def read():
        reader = csv.DictReader(csv_f, fieldnames=fieldnames)
        if strip:
            apply = operator.methodcaller("strip")
            for row in reader:
                yield {k and apply(k): (apply(v) if isinstance(v, str) else v) for k, v in row.items()}
        else:
            yield from reader

    with open(csv_path, encoding=encoding, newline="") as csv_f:
        # The official csv doc recommends opening the file with newline='' on all platforms to disable
        # universal newlines translation.

        data = itertools.islice(read(), start, stop)
        idump_into(json_path, data, encoding=encoding, **kwargs)


def idump_lines(iterable, **dumps_kwargs):
    """
    Generator yielding string lines that form a JSON array
    with the serialized elements of given iterable.
    *** Useful to reduce memory consumption ***
    - Dumps falls back to the functions: (`orjson.dumps`, `ujson.dumps`, and `json.dumps`).

    :param iterable: Iterable[dict]
    :rtype: Iterable[str]
    """

    it = iter(iterable)
    encode = functools.partial(json_dumps, **dumps_kwargs)
    indent = dumps_kwargs.get("indent")
    prefix = " " * indent if indent else ""
    yield "[\n"
    obj = next(it, None)
    if obj is not None:
        yield textwrap.indent(encode(obj), prefix)
    for obj in it:
        yield ",\n"
        yield textwrap.indent(encode(obj), prefix)
    yield "\n"
    yield "]"


def idump_into(filename, iterable, encoding=ENCODING, **kwargs):
    """
    Dump an iterable incrementally into a JSON file.
    - Use (`.gz`, `.xz`, `.bz2`) extensions to create compressed files.
    - Dumps falls back to the functions: (`orjson.dumps`, `ujson.dumps`, and `json.dumps`).

    The result will always be an array with the elements of the iterable.
    *** Useful to reduce memory consumption ***

    Examples:
        from pymince.json import idump_into

        values = ([{"key": "foo"}, {"key": "bar"}])

        idump_into("foo.json", values)     # uncompressed
        idump_into("foo.json.gz", values)  # gzip-compressed
        idump_into("foo.json.xz", values)  # lzma-compressed
        idump_into("foo.json.bz2", values) # bz2-compressed
    """

    with pymince.file.xopen(filename, mode="wt", encoding=encoding) as f:
        f.writelines(idump_lines(iterable, **kwargs))


def idump_fork(path_items, encoding=ENCODING, dump_if_empty=True, **dumps_kwargs):
    """
    Incrementally dumps different groups of elements into
    the indicated JSON file.
    *** Useful to reduce memory consumption ***

    - Use (`.gz`, `.xz`, `.bz2`) extensions to create compressed files.
    - Dumps falls back to the functions: (`orjson.dumps`, `ujson.dumps`, and `json.dumps`).

    :param Iterable[file_path, Iterable[dict]] path_items: group items by file path
    :param encoding: 'utf-8' by default.
    :param bool dump_if_empty: If false, don't create an empty file.
    :param dumps_kwargs: json.dumps kwargs.

    Examples:
        from pymince.json import idump_fork

        path_items = (
            ("num.json.gz", ({"value": 1}, {"value": 2})),
            ("num.json.gz", ({"value": 3},)),
            ("foo.json", ({"a": "1"}, {"b": 2})),
            ("baz.json", ()),
        )
        idump_fork(iter(path_items))
    """

    def get_dumper(dst):
        nothing = True
        with pymince.file.xopen(dst, mode="wt", encoding=encoding) as fd:
            write = fd.write
            write("[\n")
            try:
                while True:
                    obj = yield
                    if nothing:
                        nothing = False
                    else:
                        write(",\n")
                    write(textwrap.indent(encode(obj), prefix))
            except GeneratorExit:
                write("]") if nothing else write("\n]")

        if nothing and not dump_if_empty:
            os.unlink(dst)

    encode = functools.partial(json_dumps, **dumps_kwargs)
    indent = dumps_kwargs.get("indent")
    prefix = " " * indent if indent else ""

    dumpers = dict()
    for path, items in path_items:
        if path in dumpers:
            dumper = dumpers[path]
        else:
            dumper = get_dumper(path)
            dumper.send(None)
            dumpers[path] = dumper

        for item in items:
            dumper.send(item)
    # Cleanup
    for dumper in dumpers.values():
        dumper.close()


class JSONEncoder(json.JSONEncoder):
    """
    JSON encoder that handles additional types compared
    to `json.JSONEncoder`

    - `datetime` and `date` are serialized to strings according to the isoformat.
    - `decimal.Decimal` is serialized to a string.
    - `uuid.UUID` is serialized to a string.
    - `dataclasses.dataclass` is passed to `dataclasses.asdict`.
    - `frozenset` and `set` are serialized by ordering their values.
    """

    def default(self, obj):
        if isinstance(obj, datetime.date):
            return obj.isoformat()
        elif isinstance(obj, (frozenset, set)):
            return sorted(obj)
        elif isinstance(obj, (decimal.Decimal, uuid.UUID)):
            return str(obj)
        elif dataclasses.is_dataclass(obj):
            return dataclasses.asdict(obj)
        else:
            # Let the base class default method raise the TypeError
            return super().default(obj)



# -*- coding: utf-8 -*-

import contextlib
import json
import logging
import time


@contextlib.contextmanager
def timed_block(name, logger=None):
    """
    Log the duration of the handled context.

    Examples:
        import logging
        from pymince.logging import timed_block

        logging.basicConfig(level=logging.DEBUG)
        with timed_block("sleeping"):
            time.sleep(1)

        >>Output<<
        INFO:root:Generating [sleeping]
        DEBUG:root:Finished [sleeping in 1.002 s]
    """

    on_logger = logger or logging.getLogger()
    on_logger.info("Generating [%s]", name)
    t0 = time.time()
    try:
        yield None
    finally:
        on_logger.debug("Finished [%s in %.3f s]", name, time.time() - t0)


class StructuredFormatter(logging.Formatter):
    """
    Implementation of JSON structured logging that works
    for most handlers.

    Examples:
        import logging
        import sys
        from pymince.logging import StructuredFormatter

        # Config
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)
        formatter = StructuredFormatter('%(message)s')
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

        # Usage
        logger.debug('', {"string": "value1", "number": 1})
        logger.debug('', {"string": "value2", "number": 2})

        >>Output<<
        {"timestamp":"2022-06-17 18:37:48,789","level":"DEBUG","payload":{"string":"value1","number":1}}
        {"timestamp":"2022-06-17 18:37:48,789","level":"DEBUG","payload":{"string":"value2","number":2}}
    """

    json_dumper = json.JSONEncoder(separators=(",", ":")).encode  # Most compact form

    def format(self, record: logging.LogRecord) -> str:
        """
        Overrides parent format function.

        :param record: logging.LogRecord object
        :return: JSON string
        """

        payload = self.make_structured_dict(record)
        return self.json_dumper(payload)

    def make_structured_dict(self, record: logging.LogRecord) -> dict:
        """
        Create the dictionary that requires (json_dumper).

        :param record: logging.LogRecord object
        :return: dict
        """

        if not isinstance(record.args, dict):
            raise TypeError("Invalid logger arguments.")

        return {
            "timestamp": self.formatTime(record, self.datefmt),
            "level": record.levelname,
            "payload": record.args,
        }


class ColoredFormatter(logging.Formatter):
    """
    A class for formatting colored logs.

    Default colors:
    - DEBUG: blue
    - INFO: green
    - WARNING: yellow
    - ERROR: red
    - CRITICAL: red

    Examples:
        import logging
        from pymince.logging import ColoredFormatter

        # Config
        logger = logging.getLogger('mylog')
        logger.setLevel('DEBUG')

        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(ColoredFormatter(logging.BASIC_FORMAT))
        logger.addHandler(stream_handler)

        # Use default colors
        logger.debug("This is debug") # blue color

        # Use custom colors
        import colorama
    
        black = colorama.Fore.BLACK
        bold_black_on_green = colorama.Back.GREEN + colorama.Fore.BLACK + colorama.Style.BRIGHT
        bold_green_on_black = colorama.Back.BLACK + colorama.Fore.GREEN + colorama.Style.BRIGHT

        logger.debug("This is debug", extra={"color": black})
        logger.debug("This is debug", extra={"color": bold_black_on_green})
        logger.debug("This is debug", extra={"color": bold_green_on_black})
    """

    COLORS = {
        logging.DEBUG: "\033[94m",  # blue
        logging.INFO: "\033[92m",  # green
        logging.WARNING: "\033[93m",  # yellow
        logging.ERROR: "\033[91m",  # red
        logging.CRITICAL: "\033[1;31m",  # bold red
    }

    def format(self, record):
        formatted = super().format(record)
        if hasattr(record, "color"):
            color = record.color
        else:
            level = logging._nameToLevel[record.levelname]
            color = self.COLORS[level]
        return f"{color}{formatted}\033[0m"


class ColoredLogger:
    """
    Custom logger to generate color logs.

    Examples:
        from pymince.logging import ColoredLogger

        logger = ColoredLogger()

        # Use default colors
        logger.debug("This is debug") # blue color

        # Use custom colors
        import colorama
        logger.debug("This is debug", extra={"color": colorama.Fore.BLACK})
    """

    def __init__(self, name=None, level=logging.DEBUG, **fmt_kwargs):
        self.logger = logging.getLogger(name or __name__)
        self.logger.setLevel(level)

        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(ColoredFormatter(**fmt_kwargs))
        self.logger.addHandler(stream_handler)

    def __getattr__(self, item):
        return getattr(self.logger, item)



# -*- coding: utf-8 -*-

"""Useful functions that use dictionaries."""

import collections
import functools
import hashlib
import operator
import types

import pymince.functional
import pymince.json


def all_true_values(dictionary, keys):
    """
    Check if a dictionary has all the specified keys and if all its
    evaluated key-related values are True.

    :param dict dictionary:
    :param keys: keys sequence
    :rtype: bool

    Examples:
    from pymince.dictionary import all_true_values

    all_true_values({"a": 1, "b": 2}, ("a", "b")) # --> True
    all_true_values({"a": 1, "b": 0}, ("a", "b")) # --> False
    all_true_values({"a": 1, "b": 0}, ("a",)) # --> True
    """

    getter = operator.itemgetter(*keys)
    try:
        values = getter(dictionary)
    except KeyError:
        return False
    else:
        return all(values) if len(keys) > 1 else bool(values)


def find_leaf_value(key, dictionary):
    """
    Find leaf value in mapping.

    :param Any key: key to find
    :param dict dictionary:

    Examples:
        from pymince.dictionary import find_leaf_value

        find_leaf_value('a', {}) # --> 'a'
        find_leaf_value('a', {'a': 'b', 'b': 'c'}) # --> 'c'
        find_leaf_value('a', {'a': 'a'}) # --> 'a'
    """

    bag = set()
    while True:
        if key in dictionary:
            new_key = dictionary[key]
            if new_key == key or new_key in bag:
                break
            key = new_key
            bag.add(new_key)
        else:
            break
    return key


class DigestGetter:
    """
    Calculate a digest of a "jsonified" python dictionary.

    :param include_keys: dictionary keys to exclude
    :param exclude_keys: dictionary keys to include
    :rtype: str

    Examples:
        from pymince.dictionary import DigestGetter

        getter = DigestGetter(include_keys=("a",))
        getter({"a": 1, "b": 1}) # --> bb6cb5c68df4652941caf652a366f2d8
        getter({"a": 1}) # --> bb6cb5c68df4652941caf652a366f2d8
    """

    def __init__(self, include_keys=None, exclude_keys=None):
        if include_keys and exclude_keys:
            raise ValueError
        self.include = include_keys and frozenset(include_keys)
        self.exclude = exclude_keys and frozenset(exclude_keys)

    def __call__(self, dictionary):
        string = self.to_string(dictionary)
        return hashlib.md5(string.encode("utf-8")).hexdigest()

    @functools.cached_property
    def stringify(self):
        """
        Return a function to encode a dict into json using the most compact form.
        Dictionary keys are sorted.
        """

        return pymince.json.JSONEncoder(
            separators=(",", ":"),
            check_circular=False,
            sort_keys=True,
            ensure_ascii=False,
        ).encode

    def to_string(self, dictionary):
        # Non recursive copy.
        if self.include:
            data = {k: v for k, v in dictionary.items() if k in self.include}
        elif self.exclude:
            data = {k: v for k, v in dictionary.items() if k not in self.exclude}
        else:
            data = dictionary
        return self.stringify(data)


def frozendict(*args, **kwargs):
    """
    Returns a "MappingProxyType" from a dictionary built according to given parameters.
    Add immutability only on a first level.

    Examples:
        from pymince.dictionary import frozendict

        my_dict = frozendict(a=1, b=2)
        my_dict["a"] # --> 1
        list(my_dict.items())  # --> [("a", 1), ("b", 2)]
        my_dict["c"] = 3  # --> TypeError
    """
    return types.MappingProxyType(dict(*args, **kwargs))


def from_objects(iterable, key_getter, value_getter):
    """
    Create a new dictionary with elements generated from
    the "key_getter" and "value_getter" callbacks applied to each element of the iterable.

    :param Iterable[any] iterable:
    :param Callable key_getter:
        Dictionary keys getter.
        It is called with each element of "iterable" passing it as an argument.
    :param Callable value_getter:
        Dictionary values getter.
        It is called with each element of "iterable" passing it as an argument.

    :raise: ValueError if any generated key is duplicate.
    :rtype: dict

    Examples:
        from pymince.dictionary import from_objects

        keygetter = operator.itemgetter(0)
        valgetter = operator.itemgetter(1, 2)

        values = iter([(1, "a", "b"), (2, "a", "b")])
        from_objects(values, keygetter, valgetter) # --> {1: ('a', 'b'), 2: ('a', 'b')}
    """

    dictionary = {}
    for obj in iter(iterable):
        key = key_getter(obj)
        if key in dictionary:
            raise ValueError(key)
        else:
            dictionary[key] = value_getter(obj)
    return dictionary


def tree():
    """
    Returns a dict whose defaults are dicts.
    As suggested here: https://gist.github.com/2012250

    Examples:
        import json

        from pymince.dictionary import tree

        users = tree()
        users['user1']['username'] = 'foo'
        users['user2']['username'] = 'bar'

        print(json.dumps(users))  # {"user1": {"username": "foo"}, "user2": {"username": "bar"}}
    """

    return collections.defaultdict(tree)
